package calendar;

/**
 * Class to represent sequences (finite ordered multi-sets) of calendar entries.
 * Random access is possible in O(1).
 */
public class CalendarEntrySeq {

    protected int size = 0;
    
    //@ protected invariant contents.length >= size;
    //@ protected invariant contents.length >= 1;
    protected CalendarEntry[] contents = new CalendarEntry[1];

    //@ public invariant size() >= 0;
    //@ public invariant (\forall int i; 0 <= i && i < size(); at(i) != null);

    /**
     * @return the length of the sequence
     */
    public /*@ pure @*/ int size() {
        return size;
    }

    /**
     * @return the <code>i</code>th element of the sequence
     */
    public /*@ pure @*/ /*@ nullable @*/ CalendarEntry at(int i) {
        if ( i < 0 || i >= size ) return null;
        return contents[i];
    }

    /**
     * Insert a new element in the sequence; it is not specified in which place
     * the element will be inserted
     */
    public void insert(CalendarEntry e) {
        incSize ();
        contents[size - 1] = e;
    }

    /**
     * Remove the first occurrence of <code>iv</code> from <code>this</code>
     * sequence
     */
    public boolean removeFirst(CalendarEntry e) {
        int i = 0;
        while ( i < size && contents[i] != e )
            ++i;
        if ( i == size ) return false;
        --size;
        while ( i < size ) {
            contents[i] = contents[i + 1];
            ++i;
        }
        contents[size] = null;
        return true;
    }

    public String toString() {
        String res = "";
        int i = 0;
        while ( i < size () ) {
            res = res + at ( i ) + "\n";
            ++i;
        }
        return res;
	}
    
    /**
     * Reserve space for one further element in the sequence, and increase the
     * <code>size</code> attribute by one
     */
    /*@
      @ protected normal_behavior
      @  ensures size == \old(size) + 1;
      @  ensures (\forall int i; 0 <= i && i < \old(size);
      @                          contents[i] == \old(contents[i]));
      @  assignable contents, size;
      @*/
    protected void incSize() {
        ++size;
        if ( size > contents.length ) {
            final CalendarEntry[] oldAr = contents;
            contents = new CalendarEntry[contents.length * 2];
            // System.arraycopy ( oldAr, 0, contents, 0, oldAr.length );
            int i = 0;
            /*@
              @ loop_invariant 0 <= i && i <= oldAr.length &&
              @                (\forall int j; 0 <= j && j < i;
              @                                contents[j] == oldAr[j]);
              @ assignable i, contents[*];
              @ decreases oldAr.length - i;
              @*/
            while ( i < oldAr.length ) {
                contents[i] = oldAr[i];
                ++i;
            }
        }
    }
}
