package intervals;

/**
 * Class to represent sequences (finite ordered multi-sets) of intervals. Random
 * access is possible in O(1).
 */
public class IntervalSeq {

    protected int size = 0;
    
    //@ protected invariant contents != null;
    //@ protected invariant (\forall IntervalSeq i1, i2; i1 != i2; i1.contents != i2.contents);
    //@ protected invariant contents.length >= size;
    //@ protected invariant contents.length >= 1;
    
    //@ protected invariant (\forall int i; 0 <= i && i < size; contents[i] != null);

    //@ protected invariant !(contents instanceof interval.SimpleInterval[]);
    //@ protected invariant !(contents instanceof calendar.CalendarEntry[]);

    // protected invariant \type(Interval[]) <: \typeof(contents);

    protected /*@ nullable @*/ Interval[] contents = new Interval[1];

    //@ public invariant size >= 0;

    /**
     * @return the length of the sequence
     */
    public /*@ pure @*/ int size() {
        return size;
    }

    /**
     * @return the <code>i</code>th element of the sequence
     */
    public /*@ pure @*/ /*@ nullable @*/ Interval at(int i) {
        if ( i < 0 || i >= size ) return null;
        return contents[i];
    }

    /**
     * Insert a new element in the sequence; it is not specified in which place
     * the element will be inserted
     */
    /*@
      @ public normal_behavior
      @  ensures size == \old(size) + 1;
      @  ensures (\exists int i; 0 <= i && i < size;
      @                          contents[i] == iv
      @                          && (\forall int j; 0 <= j && j < i;
      @                                contents[j] == \old(contents[j]))
      @                          && (\forall int k; i < k && k < size;
      @                                contents[k] == \old(contents[k-1])));
      @  assignable contents, contents[*], size;
      @*/
    public void insert(Interval iv) {
        incSize ();
        contents[size - 1] = iv;
    }

    /**
     * Insert all elements of the sequence <code>seq</code> in <code>this</code>
     * sequence
     */
    public void insert(IntervalSeq seq) {
        int i = 0;
        while ( i < seq.size () ) {
            insert ( seq.at ( i ) );
            ++i;
        }
    }
  
    /**
     * Remove the first occurrence of <code>iv</code> from <code>this</code>
     * sequence
     */
    public boolean removeFirst(Interval iv) {
        int i = 0;
        while ( i < size && contents[i] != iv )
            ++i;
        if ( i == size ) return false;
        --size;
        while ( i < size ) {
            contents[i] = contents[i + 1];
            ++i;
        }
        contents[size] = null;
        return true;
    }

    /**
     * Reserve space for one further element in the sequence, and increase the
     * <code>size</code> attribute by one
     */
    /*@
      @ protected normal_behavior
      @  ensures size == \old(size) + 1;
      @  ensures (\forall int i; 0 <= i && i < \old(size);
      @                          contents[i] == \old(contents[i]));
      @  assignable contents, size;
      @*/
    protected void incSize() {
        ++size;
        if ( size > contents.length ) {
            final Interval[] oldAr = contents;
            contents = new Interval[contents.length * 2];
            // System.arraycopy ( oldAr, 0, contents, 0, oldAr.length );
            int i = 0;
            /*@
              @ loop_invariant 0 <= i && i <= oldAr.length &&
              @                (\forall int j; 0 <= j && j < i;
              @                                contents[j] == oldAr[j]);
              @ assignable i, contents[*];
              @ decreases oldAr.length - i;
              @*/
            while ( i < oldAr.length ) {
                contents[i] = oldAr[i];
                ++i;
            }
        }
    }
}
