package intervals;

/**
 * Abstract class derived from <code>IntervalSeq</code> that represents sorted
 * sequences of intervals. The sorting is realised by overwriting the
 * <code>insert</code> method, so that elements are not (as in the superclass
 * implementation) inserted as last element, but in the right place to preserve
 * the sorting. There are two concrete subclasses, sorting by the start and by
 * the end of intervals (resp.).
 */
abstract class SortedIntervalSeq extends IntervalSeq {

    /**
     * Maps intervals to the point by which they should be sorted. In the
     * subclasses, this either returns the start or the end point of an
     * interval.
     */
    protected /*@ pure @*/ abstract int getBoundary(Interval iv); 
    
    /*@ 
      @ public invariant (\forall int i; 0 <= i && i < size - 1;
      @                             getBoundary(contents[i]) <= getBoundary(contents[i+1])); 
      @*/

    /**
     * Insert a new element in the sequence
     */
    public void insert(Interval iv) {
        int i = size;
        incSize ();
        final int ivBoundary = getBoundary ( iv );

        /*@
          @ loop_invariant 0 <= i && i < size &&
          @                (\forall int k; 0 <= k && k < i; contents[k] == \old(contents[k])) &&
          @                (\forall int k; i < k && k < size; contents[k] == \old(contents[k-1])) &&
          @                (i < size - 1 ==> ivBoundary < getBoundary ( contents[i+1] ));
          @ decreases i;
          @ assignable contents[*], i;
          @*/
        while ( i > 0 && ivBoundary < getBoundary ( contents[i-1] ) ) {
            contents[i] = contents[i - 1];
            --i;
        }
        contents[i] = iv;
    }

    /**
     * Insert all elements <code>x</code> of this sequence with
     * <code>getBoundary(x)<=p</code> into <code>seq</code>.
     */
    public void collectLeq(IntervalSeq seq, int p) {
        int i = 0;
        while ( i < size && getBoundary ( contents[i] ) <= p ) {
            seq.insert ( contents[i] );
            ++i;
        }
    }

    /**
     * Insert all elements <code>x</code> of this sequence with
     * <code>getBoundary(x)>=p</code> into <code>seq</code>.
     */
    public void collectGeq(IntervalSeq seq, int p) {
        int i = size - 1;
        while ( i >= 0 && getBoundary ( contents[i] ) >= p ) {
            seq.insert ( contents[i] );
            --i;
        }
    }
}
